from typing import Callable, Protocol, TextIO, ContextManager, TypeVar, Any, Optional, Generic, overload

T0 = TypeVar('T0')
T1 = TypeVar('T1')
T2 = TypeVar('T2')
T3 = TypeVar('T3')
T4 = TypeVar('T4')
T5 = TypeVar('T5')

T0_contra = TypeVar('T0_contra', contravariant=True)
T1_contra = TypeVar('T1_contra', contravariant=True)
T2_contra = TypeVar('T2_contra', contravariant=True)
T3_contra = TypeVar('T3_contra', contravariant=True)
T4_contra = TypeVar('T4_contra', contravariant=True)
T5_contra = TypeVar('T5_contra', contravariant=True)

class F0(Protocol):
  def __call__(self, *,
    width: int) -> str: ...

class F1(Protocol[T0_contra]):
  def __call__(self, __arg0: T0_contra,
    width: int) -> str: ...

class F2(Protocol[T0_contra,T1_contra]):
  def __call__(self, __arg0: T0_contra, __arg1: T1_contra, *,
    width: int) -> str: ...

class F3(Protocol[T0_contra,T1_contra,T2_contra]):
  def __call__(self, __arg0: T0_contra, __arg1: T1_contra, __arg2: T2_contra, *,
    width: int) -> str: ...

class F4(Protocol[T0_contra,T1_contra,T2_contra,T3_contra]):
  def __call__(self, __arg0: T0_contra, __arg1: T1_contra, __arg2: T2_contra, __arg3: T3_contra, *,
    width: int) -> str: ...

class F5(Protocol[T0_contra,T1_contra,T2_contra,T3_contra,T4_contra]):
  def __call__(self, __arg0: T0_contra, __arg1: T1_contra, __arg2: T2_contra, __arg3: T3_contra, __arg4: T4_contra, *,
    width: int) -> str: ...

class F6(Protocol[T0_contra,T1_contra,T2_contra,T3_contra,T4_contra,T5_contra]):
  def __call__(self, __arg0: T0_contra, __arg1: T1_contra, __arg2: T2_contra, __arg3: T3_contra, __arg4: T4_contra, __arg5: T5_contra, *,
    width: int) -> str: ...

class Fany(Protocol):
  def __call__(*args: Any,
    width: int) -> str: ...

@overload
def BottomBar(*,
  format: F0, output: TextIO = ..., interval: Optional[float] = ...) -> ContextManager[Callable[[], None]]: ...

@overload
def BottomBar(__arg0: T0, *,
  format: F1[T0], output: TextIO = ..., interval: Optional[float] = ...) -> ContextManager[Callable[[T0], None]]: ...

@overload
def BottomBar(__arg0: T0, __arg1: T1, *,
  format: F2[T0, T1], output: TextIO = ..., interval: Optional[float] = ...) -> ContextManager[Callable[[T0, T1], None]]: ...

@overload
def BottomBar(__arg0: T0, __arg1: T1, __arg2: T2, *,
  format: F3[T0, T1, T2], output: TextIO = ..., interval: Optional[float] = ...) -> ContextManager[Callable[[T0, T1, T2], None]]: ...

@overload
def BottomBar(__arg0: T0, __arg1: T1, __arg2: T2, __arg3: T3, *,
  format: F4[T0, T1, T2, T3], output: TextIO = ..., interval: Optional[float] = ...) -> ContextManager[Callable[[T0, T1, T2, T3], None]]: ...

@overload
def BottomBar(__arg0: T0, __arg1: T1, __arg2: T2, __arg3: T3, __arg4: T4, *,
  format: F5[T0, T1, T2, T3, T4], output: TextIO = ..., interval: Optional[float] = ...) -> ContextManager[Callable[[T0, T1, T2, T3, T4], None]]: ...

@overload
def BottomBar(__arg0: T0, __arg1: T1, __arg2: T2, __arg3: T3, __arg4: T4, __arg5: T5, *,
  format: F6[T0, T1, T2, T3, T4, T5], output: TextIO = ..., interval: Optional[float] = ...) -> ContextManager[Callable[[T0, T1, T2, T3, T4, T5], None]]: ...

@overload
def BottomBar(*args: Any,
  format: Fany, output: TextIO = ..., interval: Optional[float] = ...) -> ContextManager[Callable[..., None]]: ...

@overload
def BottomBar(*args: Any,
  output: TextIO = ..., interval: Optional[float] = ...) -> ContextManager[Callable[..., None]]: ...

# vim:sw=2:sts=2:et



# Copyright (c) 2020 Evalf
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

'''Print and maintain a status line at the bottom of a VT100 terminal.'''

__version__ = '1.0'


import sys, os, signal, threading, time


class _ontime(threading.Thread):
  def __init__(self, interval, callback, *args):
    super().__init__()
    self.interval = interval
    self.callback = callback
    self.args = args
    self.stopped = threading.Event()
    self.start()
  def run(self):
    interval = abs(self.interval)
    precise = self.interval > 0
    while not self.stopped.wait(interval - (precise and time.perf_counter() % interval)):
      self.callback(*self.args)
  def close(self):
    self.stopped.set()
    self.join()


class _onevent:
  def __init__(self, signalnum, callback, *args):
    self.callback = callback
    self.args = args
    self.signalnum = signalnum
    self.oldhandler = signal.signal(signalnum, self.handler)
  def handler(self, sig, tb):
    self.callback(*self.args)
  def close(self):
    signal.signal(self.signalnum, self.oldhandler)


def _format(*args, width):
  return ' '.join(map(str, args))


class BottomBar:
  '''Bar drawing context.

  The BottomBar context is created with a single string argument, which is
  displayed for the duration that the context remains entered.

  >>> with BottomBar('bar text'):
  ...     print('regular output')

  To change the bar text without having to create a new context, the BottomBar
  context returns an update method upon entry.

  >>> with bottombar.BottomBar('bar text') as bb:
  ...     bb('new bar text')

  The optional keyword argument `format` determines how the bottom bar is drawn.
  In addition to any number of positional arguments, this callable must also be
  able to receive the keyword argument `width` for the present number of columns.
  In the following example this argument is used to center the bar text.

  >>> def center(text, width):
  ...     return text.center(width)
  >>> with bottombar.BottomBar('bar text', format=center) as bb:
  ...     print('regular output')

  The optional keyword argument `interval` causes BottomBar to automatically call
  its formatter at set intervals. This is useful for formatters that (partly)
  generate their own content. For example, the following adds a right-aligned
  clock to the bar text, updated every second:

  >>> import time
  >>> def myformat(text, width):
  ...     return text + time.ctime().rjust(width - len(text))
  >>> with bottombar.BottomBar('bar text', format=myformat, interval=1) as bb:
  ...     bb('new bar text')'''

  def __init__(self, *args, output=sys.stdout, format=_format, interval=None):
    self.args = args
    self.encoding = output.encoding
    self.fileno = output.fileno()
    self.format = format
    self.interval = interval
    self.size = None
    self.handles = []

  def __enter__(self):
    if self.size is not None:
      raise RuntimeError('BottomBar is not reentrant')
    self.redraw(True)
    if hasattr(signal, 'SIGWINCH'):
      self.handles.append(_onevent(signal.SIGWINCH, self.redraw, False))
    elif not self.interval:
      self.handles.append(_ontime(-1, self.redraw, False))
    if self.interval:
      self.handles.append(_ontime(self.interval, self.redraw, True))
    return self.update

  def write(self, data):
    while data:
      n = os.write(self.fileno, data)
      data = data[n:]

  def update(self, *args):
    self.args = args
    self.redraw(True)

  def redraw(self, force):
    size = os.get_terminal_size(self.fileno)
    if size != self.size:
      self.write(
        b'\0337' # save cursor and attributes
        b'\033[r' # reset scroll region (moves cursor)
        b'\0338' # restore cursor and attributes
        b'\033D' # move/scroll down
        b'\033M' # move up
        b'\0337' # save cursor and attributes
        b'\033[1;%dr' # set scroll region
        b'\0338' # restore cursor and attributes
        % (size.lines-1))
      self.size = size
      force = True
    if force:
      self.write(
        b'\0337' # save cursor and attributes
        b'\033[%d;1H' # move cursor to bottom row, first column
        b'\033[2K' # clear entire line
        b'\033[?7l' # disable line wrap
        b'\033[0m' # clear attributes
        b'%s' # print bar
        b'\033[?7h' # enable line wrap
        b'\0338' # restore cursor and attributes
        % (size.lines, self.format(*self.args, width=size.columns).encode(self.encoding)))

  def __exit__(self, *exc):
    if self.size is None:
      raise RuntimeError('BottomBar has not yet been entered')
    while self.handles:
      self.handles.pop().close()
    self.write(
      b'\0337' # save cursor position
      b'\033[%d;1H' # move cursor to bottom row, first column
      b'\033[K' # clear entire line
      b'\033[r' # reset scroll region
      b'\0338' # restore cursor position
      % self.size.lines)
    self.size = None


# vim:sw=2:sts=2:et



import bottombar, os, time

print('\033[{};1H'.format(os.get_terminal_size().lines))
print('====================')
print('BOTTOMBAR UNIT TESTS')
print('--------------------')
print('1. basic terminal tests')
input('   a. check that this question shows at the very last line')
input('   b. check that the the last word shows in \033[0;33myellow\033[0m')
input('   c. check that this box [ ] contains the blinking cursor\033[31D')
print('2. basic bottombar tests')
with bottombar.BottomBar('foo...') as bb:
  input('   a. check that the bottom line shows the text "foo..."')
  bb('bar!')
  input('   b. check that the bottom line shows "bar!" (and nothing more)')
  input('   c. check that this {}long sentence wraps without overwriting "bar!"'.format('very ' * 40))
  bb('abc', 123)
  input('   d. check that the bottom line shows "abc 123"')
input('   e. check that the bar disappeared')
print('3. line wrapping tests')
with bottombar.BottomBar('this is a very long bar; ' * 99) as bb:
  input('   a. check that the line below is truncated without wrapping')
  print('   b. check that \033[0;33m[this entire ', end='', flush=True)
  bb('white bar')
  input('box]\033[0m is yellow but the bar is white')
  print('   c. check that \033[0;33m[this entire ', end='', flush=True)
  bb('\033[0;34mblue bar')
  input('box]\033[0m is yellow but the bar is blue')
  print('   d. check that \033[0;33m[this entire ', end='', flush=True)
input('box]\033[0m is yellow and the bar disappeared')
print('4. resize tests')
with bottombar.BottomBar('left', 'right', format=lambda a, b, width: a.ljust(width-len(b),'-') + b) as bb:
  input('   a. check that the bottom line shows "left--(...)--right" over the entire length')
  input('   b. change the window width; check that "left" and "right" move along')
  input('   c. change the window height; check that the bar remains at the bottom')
  input('   e. check that the blinking cursor stays in this box [ ] upon resize\033[14D')
print('5. update tests')
def format(text: str, width: int) -> str:
  t = time.time()
  time.sleep(.123)
  return '{0.tm_hour:02d}:{0.tm_min:02d}:{1:06.3f} {2}'.format(time.localtime(t), t%60, text)
with bottombar.BottomBar('updating approximately every second', format=format, interval=-1) as bb:
  input('   a. check that the time in the bar updates approximately every second')
  input('   b. check that the milliseconds change with every update')
with bottombar.BottomBar('updating precisely every second', format=format, interval=+1) as bb:
  input('   c. check that the milliseconds remains fixed with every update')
  bb('the end.')
  input('   d. check that the text after the time changed to "the end."')
print('--------------------')
